# Copyright 2008 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2
# Author: Michael Haubenwallner <haubi@gentoo.org>
#
# Utilities for dynamic link libraries (aka. shared objects) on AIX.
#

PS4='(${LINENO})+ '

AIXDLL_ACTION=${0##*/}
[[ ${AIXDLL_ACTION} == 'aixdll-'* ]] && AIXDLL_ACTION=${AIXDLL_ACTION#aixdll-} || AIXDLL_ACTION=

_aixdll_die() {
	if type die >&/dev/null; then
		[[ $# -gt 0 ]] && die "${AIXDLL_ACTION}: $@"
		die
	else
		[[ $# -gt 0 ]] && echo "${AIXDLL_ACTION}: $@" >&2
		exit 1
	fi
}

_aixdll_cutoff() {
	local line=$1
	local minlen=$2
	local maxcol=$((${#line}-${minlen}))
	local tmpline value
	if [[ ${line:$((maxcol-1)):1} != ' ' ]]; then
		tmpline=${line:0:${maxcol}}
		tmpline=${tmpline% *}
		maxcol=$((${#tmpline}+1))
	fi
	value=${line:${maxcol}}
	echo ${value}
	echo "${line%${value}}"
}

_aixdll_read_RUNPATH_DEPLIBS() {
	# Reads stdin up to the first empty line.
	# The first line on stdin must be:
	#     "INDEX PATH BASE MEMBER "
	# The blank-counts in this line are used to parse the data.
	#
	# Output:
	# Two lines going to stdout:
	# "/path1:/path2"
	# "'lib1.a(shr1.o)' '/path/lib2.a(shr2.o)' 'lib3.so'"
	local line
	local runpath= deplibs=
	local haveHeader=false
	local dynMEMBER dynBASE dynPATH

	# global constants to calculate once only
	declare -i PATHlen=0 BASElen=0 MEMBERlen=0
	declare -i PATHmin=0 BASEmin=0 MEMBERmin=0

	while IFS= read line
	do
		# empty line marks end of ImportFileStrings section
		[[ ${line} ]] || break
		if [[ ${line} == "INDEX "* ]]; then
			haveHeader=true
			if [[ ${PATHmin} -eq 0 ]]; then
				MEMBERlen=${#line}
				line=${line%MEMBER *}
				MEMBERmin=${#line}
				MEMBERlen=$((MEMBERlen-MEMBERmin))
				BASElen=${MEMBERmin}
				line=${line%BASE *}
				BASEmin=${#line}
				BASElen=$((BASElen-BASEmin))
				PATHlen=${BASEmin}
				line=${line%PATH *}
				PATHmin=${#line}
				PATHlen=$((PATHlen-PATHmin))
			fi
			continue
		fi
		${haveHeader} || continue

		{ IFS= read dynMEMBER; IFS= read line; } <<-EOC
			$(_aixdll_cutoff "${line}" ${MEMBERlen})
		EOC
		{ IFS= read dynBASE; IFS= read line; } <<-EOC
			$(_aixdll_cutoff "${line}" ${BASElen})
		EOC
		{ IFS= read dynPATH; IFS= read line; } <<-EOC
			$(_aixdll_cutoff "${line}" ${PATHlen})
		EOC

		if [[ ! ${dynBASE} && ${dynPATH} ]]; then
			runpath=${dynPATH}
			continue
		fi

		[[ ${dynBASE} && ${dynBASE} != .. && ${dynBASE} != . ]] || continue

		deplibs="${deplibs} '${dynPATH%/}${dynPATH:+/}${dynBASE}${dynMEMBER:+(}${dynMEMBER}${dynMEMBER:+)}'"
	done

	echo "${runpath}"
	echo "${deplibs}"
}

aixdll_query() {
	# Uses /usr/bin/dump to extract information from object files
	# or archive libraries.
	#
	# First argument:
	#   either
	#     "/bin/true"
	#        queries binary information from file.
	#   or
	#     "/lib/libc.a"
	#        queries binary information for each archive member.
	#   or
	#     "/lib/libc.a(shr.o)"
	#        queries binary information for specified archive members,
	#        archive-members can contain shell-wildcards (*?).
	# Further arguments:
	#   any combination of
	#     FILE MEMBER SHROBJ FLAGS FORMAT RUNPATH DEPLIBS
	#
	# Output on stdout:
	#   One line for each file or archive member, to be evaluated in
	#   another shell environment, setting at least these variables:
	#     FILE='...';MEMBER='...';SHROBJ=[true|false]
	#   The format for each variable is
	#   either (for '/bin/true')
	#     FILE='/bin/true'
	#     MEMBER=''
	#     FLAGS=
	#     FORMAT=
	#     RUNPATH=
	#     DEPLIBS=
	#     SHROBJ=false
	#     FLAGS=' EXEC DYNLOAD SHROBJ DEP__SYSTEM '
	#     FORMAT='32-bit XCOFF'
	#     RUNPATH='/usr/lib:/lib:/usr/lpp/xlC/lib'
	#     DEPLIBS=" 'libc.a(shr.o)'"
	#   or (for '/lib/libc.a')
	#     FILE='/lib/libc.a'
	#     MEMBER='ptrgl_64_64.o'
	#     FLAGS=
	#     FORMAT=
	#     RUNPATH=
	#     DEPLIBS=
	#     SHROBJ=false
	#     FLAGS=' DEP_SYSTEM '
	#     FORMAT='64-bit XCOFF'
	#    <newline for each archive member>
	#     FILE='/lib/libc.a'
	#     MEMBER='shr.o'
	#     FLAGS=
	#     FORMAT=
	#     RUNPATH=
	#     DEPLIBS=
	#     SHROBJ=true
	#     FLAGS=' EXEC DYNLOAD SHROBJ DEP__SYSTEM '
	#     FORMAT='32-bit XCOFF'
	#     RUNPATH='/usr/lib:/lib'
	#     DEPLIBS=" '/unix' 'libcrypt.a(shr.o)'"
	#    <newline>
	#   or (for '/lib/libc.a(shr*)')
	#     FILE='/lib/libc.a'
	#     MEMBER='shr.o'
	#     FLAGS=
	#     FORMAT=
	#     RUNPATH=
	#     DEPLIBS=
	#     SHROBJ=true
	#     FLAGS=' EXEC DYNLOAD SHROBJ DEP__SYSTEM '
	#     FORMAT='32-bit XCOFF'
	#     RUNPATH='/usr/lib:/lib'
	#     DEPLIBS=" '/unix' 'libcrypt.a(shr.o)'"
	#    <newline>
	#     FILE='/lib/libc.a'
	#     MEMBER='shr_64.o'
	#     FLAGS=
	#     FORMAT=
	#     RUNPATH=
	#     DEPLIBS=
	#     SHROBJ=true
	#     FLAGS=' EXEC DYNLOAD SHROBJ DEP__SYSTEM '
	#     FORMAT='64-bit XCOFF'
	#     RUNPATH='/usr/lib:/lib'
	#     DEPLIBS=" '/unix' 'libcrypt.a(shr_64.o)'"
	#    <newline>

	aixdll-query "$@"
}

aixdll_find_unprepared() {
	find "$1" -type f -name 'lib*.a' -print
}

aixdll_find_prepared() {
	find "$1" -type d -name 'lib*.a.d' -print
}

aixdll_is_prepared() {
	local what=$1
	[[ ${what##*/} == lib*.a.d ]]
}

aixdll_get_preparedir_for_lib() {
	echo "${1%.d}.d"
}

aixdll_get_lib_for_preparedir() {
	echo "${1%.d}"
}

aixdll_prepare_for_merge() {
	# $0 target-library [source-library]
	local libsrc=${2-${1}}
	local libad=$(aixdll_get_preparedir_for_lib "${1}")
	mkdir -p "${libad}" || _aixdll_die "Cannot create ${libad}"
	pushd "${libad}" >/dev/null || _aixdll_die "Cannot cd to ${libad}"
	local status
	/usr/ccs/bin/ar -X32_64 -x "${libsrc}"
	status=$?
	popd >/dev/null || _aixdll_die "Cannot cd back from ${libad}"
	if [[ ${status} -ne 0 ]]; then
		rmdir "${libad}" >&/dev/null
		_aixdll_die "Cannot un-ar ${libsrc}"
	fi
	true
}

aixdll_finish_merge_single() {
	local arg keepdir
	local libad
	for arg
	do
		case ${arg} in
		--keepdir=*)
			keepdir=${arg#*=}
			continue
			;;
		*)
			libad=$(aixdll_get_preparedir_for_lib "${arg}")
			;;
		esac
	done

	: ${keepdir:=true}

	aixdll_is_prepared "${libad}" || return 0
	local liba=$(aixdll_get_lib_for_preparedir "${libad}")
	local f

	if [[ -d ${libad} && $(ls -A "${libad}/") ]]; then
		pushd "${libad}" >/dev/null || _aixdll_die "Cannot cd to ${libad}"
		rm -f "./${liba##*/}.new" || _aixdll_die "Cannot remove ${liba##*/}.new"
		/usr/ccs/bin/ar -coqszvl -X 32_64 "./${liba##*/}.new" ./* || _aixdll_die "Cannot recreate ${liba}"
		mv -f "./${liba##*/}.new" "${liba}" || _aixdll_die "Cannot move ${liba##*/}.new to ${liba}"
		popd >/dev/null || _aixdll_die "Cannot cd back from ${libad}"
	elif [[ -f ${liba} ]]; then
		rm -f "${liba}" || _aixdll_die "Cannot prune '${liba}'"
		keepdir=false
	fi

	if [[ -d ${libad} ]] && ! ${keepdir}; then
		pushd "${libad}" >/dev/null || _aixdll_die "Cannot cd to ${libad}"
		rm -f ./* || _aixdll_die "Cannot prune '${libad}/*'"
		popd >/dev/null || _aixdll_die "Cannot cd back from ${libad}"
		rmdir "${libad}" || _aixdll_die "Cannot prune '${libad}'"
	fi
	true
}

aixdll_finish_merge() {
	local arg status keepdir
	for arg
	do
		[[ ${arg} == --keepdir=* ]] && keepdir=${arg#*=}
	done

	for arg
	do
		[[ ${arg} != --* ]] || continue
		( aixdll_finish_merge_single --keepdir=${keepdir} "${arg}" ) || status=$?
	done
	return ${status}
}

aixdll_merge_runtime() {
	[[ $# -gt 0 ]] || return 0
	local arg target current curdir libad finish keepdir
	local curdir=$(pwd)
	for arg
	do
		case "${arg}" in
		--target=*) target=${arg#*=} ;;
		--current=*) current=${arg#*=} ;;
		--finish=*) finish=${arg#*=} ;;
		--keepdir=*) keepdir=${arg#*=} ;;
		--*) _aixdll_die "Invalid argument ${arg}" ;;
		esac
	done
	: ${finish:=true}
	[[ ${target} ]] || _aixdll_die "Missing --target"
	[[ ${target} == /* ]] || target=${curdir}/${target}
	[[ ${current} && ${current} != /* ]] && current=${curdir}/${current}
	libad=$(aixdll_get_preparedir_for_lib "${target}")
	local line status
	local FILE MEMBER SHROBJ FLAGS
	for arg
	do
		[[ ${arg} == '--'* ]] && continue
		[[ ${arg} == /* ]] || arg=${curdir}/${arg}

		while IFS= read line
		do
			[[ ${line} == *SHROBJ=* ]] || continue
			eval "${line}"
			${SHROBJ} || continue
			mkdir -p "${libad}" || _aixdll_die "Cannot create '${libad}'"
			pushd "${libad}" >/dev/null || _aixdll_die "Cannot cd to ${libad}"
			/usr/ccs/bin/ar -X32_64 -x "${FILE}" "${MEMBER}"
			status=$?
			if [[ ${status} -ne 0 ]]; then
				popd >/dev/null || _aixdll_die "Cannot cd back from ${libad}"
				rmdir "${libad}" >&/dev/null
				_aixdll_die "Cannot extract '${MEMBER}' of '${FILE}'"
			fi
			if [[ ${arg} != "${current}" ]]; then
				[[ ${FLAGS} == *" LOADONLY "* ]] ||
				/usr/ccs/bin/strip -e "./${MEMBER}" ||
				_aixdll_die "Cannot set flag LOADONLY for ${MEMBER}"
			else
				[[ ${FLAGS} != *" LOADONLY "* ]] ||
				/usr/ccs/bin/strip -E "./${MEMBER}" ||
				_aixdll_die "Cannot drop flag LOADONLY from ${MEMBER}"
			fi
			popd >/dev/null || _aixdll_die "Cannot cd back from ${libad}"
		done <<-EOF
			$(aixdll_query "${arg}" FILE MEMBER SHROBJ FLAGS)
		EOF
		:
	done
	${finish} || return 0
	aixdll_finish_merge --keepdir=${keepdir} "${libad}"
}

aixdll_help() {
	cat <<EOF
Known actions:
	--query
	--find-unprepared
	--find-prepared
	--is-prepared
	--prepare-for-merge
	--finish-merge
	--merge-runtime
EOF
}

aixdll_main() {
	: ${AIXDLL_ACTION:=help}
	while [[ ${1:+set} ]]
	do
		case ${1} in
		--help | \
		--query | \
		--find-unprepared | \
		--find-prepared | \
		--is-prepared | \
		--prepare-for-merge | \
		--finish-merge | \
		--merge-runtime )
			AIXDLL_ACTION=${1#--}
			;;
		--debug) set -x ;;
		*) break ;;
		esac
		shift
	done

	aixdll_${AIXDLL_ACTION//-/_} "$@"
}
