/* Copyright 2008 Gentoo Foundation
 * Distributed under the terms of the GNU General Public License v2
 * Author: Michael Haubenwallner <haubi@gentoo.org>
 */

#ifndef _aixdll_ObjectKey_h
#define _aixdll_ObjectKey_h

#include <string>

#include "smartptr.h"

namespace aixdll {

class ObjectPath : public std::string, virtual public RefCounted
{
public:
	~ObjectPath() {}

	ObjectPath(std::string const& path)
		: std::string(path)
	{
		if (length() > 0 && operator[](length() - 1) != '/') {
			append("/");
		}
	}

	std::string const& asIdentifier() const { return *this; }
	std::string const& string() const { return *this; }
};

typedef SmartPtr<ObjectPath> ObjectPathPtr;

class ObjectName : virtual public RefCounted {
public:
	~ObjectName() {}

	ObjectName(std::string const& base, std::string const& member)
		: base_(base)
		, member_(member)
	{}

	std::string const& base() const { return base_; }
	std::string const& member() const { return member_; }

	bool operator< (ObjectName const& that) const
	{
		if (base_ < that.base_) return true;
		if (member_ < that.member_) return true;
		return false;
	}

	std::string asIdentifier() const
	{
		std::string identifier = base_;
		if (member_.length() > 0) {
			identifier += "[";
			identifier += member_;
			identifier += "]";
		}
		return identifier;
	}

private:
	std::string base_;
	std::string member_;
};

typedef SmartPtr<ObjectName> ObjectNamePtr;

static void splitIdentifier(std::string const& identifier,
		ObjectPathPtr &objPath, ObjectNamePtr &objName)
{
	objPath.forget();
	objName.forget();

	std::string::const_iterator end = identifier.end();
	if (end == identifier.begin()) {
		return;
	}

	std::string path, base, member;

	std::string::const_iterator start = end;
	--start;
	if (*start == ')' || *start == ']') {
		std::string::value_type starter = (*start == ')') ? '(' : '[';
		end = start;
		while(start > identifier.begin()) {
			if (*start == starter) {
				member.assign(start+1, end);
				end = start;
				start--;
				break;
			}
			start--;
		}
	}
	while(start > identifier.begin() && *start != '/') {
		start--;
	}

	if (*start == '/') {
		start++;
	}
	base.assign(start, end);

	if (start > identifier.begin()) {
		path.assign(identifier.begin(), start);
	}

	if (path.length() > 0) {
		objPath.eat(new ObjectPath(path));
	}
	objName.eat(new ObjectName(base, member));
}

class ObjectKey : virtual public RefCounted {
public:
	~ObjectKey() {}

	ObjectKey(std::string const& identifier)
		: path_()
		, name_()
	{
		splitIdentifier(identifier, path_, name_);
	}

	ObjectKey(ObjectPathPtr const& path, ObjectNamePtr const& name)
		: path_(path)
		, name_(name)
	{}

	ObjectPathPtr const& path() const { return path_; }
	ObjectNamePtr const& name() const { return name_; }

	std::string asIdentifier() const
	{
		return path_ ? path_->asIdentifier() + name_->asIdentifier() : name_->asIdentifier();
	}

	bool operator<(ObjectKey const& that) const
	{
		if (path_ < that.path_) return true;
		if (name_ < that.name_) return true;
		return false;
	}
private:
	ObjectPathPtr path_;
	ObjectNamePtr name_;
};

typedef SmartPtr<ObjectKey> ObjectKeyPtr;

}

#endif /* _aixdll_ObjectKey_h */
