/* Copyright 2008 Gentoo Foundation
 * Distributed under the terms of the GNU General Public License v2
 * Author: Michael Haubenwallner <haubi@gentoo.org>
 */

#include <errno.h>
#include <string.h>
#include <stdexcept>

#include "StaticObject.h"
#include "Executable.h"

#include "ObjectReader.h"

namespace aixdll {

static bool readLine(FILE *file, std::string& line)
{
	char linebuf[1024];

	line = "";

	do {
		/* fgets always stores '\0' into the last byte read.
		 * If the byte before the last possible one is either
		 * unchanged or newline, we've read a whole line.
		 */
		linebuf[sizeof(linebuf)-2] = '\0';
		char *read = fgets(linebuf, sizeof(linebuf), file);
		if (read == NULL) {
			if (feof(file)) {
				return false;
			}
			throw std::runtime_error(strerror(errno));
		}
		line += read;
	} while(linebuf[sizeof(linebuf)-2] != '\0' && linebuf[sizeof(linebuf)-2] != '\n');

	while(line.length() > 0
	   && (line[line.length()-1] == '\n' || line[line.length()-1] == '\r')
	) {
		line.erase(line.length()-1);
	}
	return true;
}

static void cutoff(std::string& line, std::string::size_type minlen, std::string& out)
{
	if (line.length() < 1 || line.length() <= minlen) {
		return;
	}
	std::string::size_type blank = line.rfind(" ", line.length() - minlen - 1);
	if (blank == line.npos) {
		return;
	}
	blank++;
	out.assign(line, blank, line.length() - blank);
	line.erase(blank, line.length() - blank);
	std::string::iterator end = out.end();
	while(--end >= out.begin()) {
		if (*end != ' ') {
			break;
		}
		end = out.erase(end);
	}
}

class ObjectReader::Dumper {
public:
	Dumper() : file_(0) {}

	~Dumper()
	{
		close();
	}

	bool open(std::string const& filename)
	{
		std::string dumpcommand = "/usr/bin/dump -Hov -X32_64 '";
		dumpcommand += filename;
		dumpcommand += "' 2>/dev/null";
		file_ = popen(dumpcommand.c_str(), "r");
		return file_ ? true : false;
	}

	int close()
	{
		if (file_) {
			return pclose(file_);
		}
		return -1;
	}

	FILE* file() { return file_; }
private:
	FILE *file_;
};

void ObjectReader::parseRunpathDeplibs(Dumper *dumper, SharedObjectPtr &object) const
{
	std::string line;
	bool haveHeader = false;

	std::string::size_type MEMBERlen = 0, MEMBERmin = 0;
	std::string::size_type BASElen = 0, BASEmin = 0;
	std::string::size_type PATHlen = 0, PATHmin = 0;

	std::string path, base, member;

	while(readLine(dumper->file(), line)) {
		if (line.empty()) {
			break;
		}
		if (line.find("INDEX ") == 0) {
			haveHeader = true;
			MEMBERmin = line.rfind("MEMBER ");
			MEMBERlen = line.length() - MEMBERmin;
			BASEmin = line.rfind("BASE ", MEMBERmin);
			BASElen = MEMBERmin - BASEmin;
			PATHmin = line.rfind("PATH ", BASEmin);
			PATHlen = BASEmin - PATHmin;
			continue;
		}
		if (! haveHeader) continue;

		cutoff(line, MEMBERlen, member);
		cutoff(line, BASElen, base);
		cutoff(line, PATHlen, path);

		if (base == "." || base == "..") {
			continue;
		}

		if (base.length() == 0) {
			object->setRunpath(SearchPathPtr(new SearchPath(path)));
		} else
		if (path.length() > 0) {
			object->addDependency
			(	DependencyPtr(new Dependency
				(	ObjectKeyPtr(
						new ObjectKey
					(	ObjectPathPtr(new ObjectPath(path))
					  ,	ObjectNamePtr(new ObjectName(base, member))
					)
				))
			));
		} else {
			object->addDependency
			(	DependencyPtr(new Dependency
				(	ObjectNamePtr(new ObjectName(base, member)))
			));
		}
	}
}

int ObjectReader::readObjects(std::string const& filename
		, std::string const& identifier
		, ObjectContainer& out
)
{
	int rv = 0;
	/* filename-based */
	ObjectPathPtr fileObjPath;
	ObjectNamePtr fileObjName;
	std::string filePath;

	std::string base; /* same for filename and identifier */
	std::string member; /* same for filename and identifier */

	/* identifier-based */
	ObjectPathPtr idObjPath;
	ObjectNamePtr idObjName;
	std::string idPath;

	/* currently read */
	ObjectPathPtr objPath;
	ObjectNamePtr objName;
	std::string currMember;

	splitIdentifier(filename, fileObjPath, fileObjName);
	splitIdentifier(identifier, idObjPath, idObjName);

	if (idObjPath) {
		idPath = idObjPath->string();
	}

	if (fileObjPath) {
		filePath = fileObjPath->string();
	}
	if (fileObjName && fileObjName->base().length() > 0) {
		base = fileObjName->base();
	}
	if (fileObjName && fileObjName->member().length() > 0) {
		member = fileObjName->member();
	}

	if (filePath.length() == 0) {
		filePath = "./";
	}

	LibraryMemberCache::iterator cachedLib;
	if (member.length() > 0) {
		member.insert(0, "[");
		member.append("]");
		cachedLib = libraryMemberCache_.end();
	} else {
		cachedLib = libraryMemberCache_.find(idPath+base);
	}

	if (objectCache_.find(idPath + base + member) != objectCache_.end()) {
		out.push_back(objectCache_[idPath + base + member]);
		return 1;
	}

	if (cachedLib != libraryMemberCache_.end()) {
		for(LibraryMemberNameContainer::iterator it = cachedLib->second.begin()
		  ; it != cachedLib->second.end()
		  ; ++it
		) {
			out.push_back(objectCache_[idPath + base + *it]);
			++ rv;
		}
		return rv;
	}

	if (brokenFilesCache_.find(filePath + base) != brokenFilesCache_.end()) {
		return 0;
	}

	Dumper dumper;

	if (! dumper.open(filePath + base)) {
		brokenFilesCache_[filePath + base] = true;
		return 0;
	}

	std::string line;

	enum { NoSection, ImportFileStrings, ObjectModuleHeader } section = NoSection;

	ObjectKeyPtr objectKey;
	std::string flags, format;

	ObjectPtr objectPtr;
	SharedObjectPtr sharedObject;
	StaticObjectPtr staticObject;

	while(readLine(dumper.file(), line)) {
		if (! line.empty() && line[line.length()-1] == ':') {
			if (objectPtr != NULL) {
				objectCache_[idPath + base + currMember] = objectPtr;
				if (currMember.length() > 0) {
					libraryMemberCache_[idPath + base].push_back(currMember);
				}
				if (member.length() == 0 || member == currMember) {
					out.push_back(objectPtr);
					++ rv;
				}
				objectPtr.forget();
				sharedObject.forget();
				staticObject.forget();
			}

			currMember.clear();

			splitIdentifier(std::string(&line[0], &line[line.length()-1]),
						objPath, objName);

			currMember = objName->member();

			if (currMember.length() > 0) {
				currMember.insert(0, "[");
				currMember.append("]");
			}

			if (objName->base() != base || objPath->string() != filePath) {
				/* Huh? */
				throw std::runtime_error("unexpected dump output");
			}
			objectKey.eat(new ObjectKey(idObjPath, objName));
			flags="";
			format="";
			continue;
		}

		if (line.empty()) {
			section = NoSection;
			continue;
		}

		if (! objectPtr && flags.length() > 0 && format.length() > 0) {
			if (flags.find("SHROBJ") != flags.npos) {
				sharedObject.eat(new SharedObject(objectKey, flags, format));
				objectPtr = sharedObject;
			} else
			if (flags.find("EXEC") != flags.npos) {
				sharedObject.eat(new Executable(objectKey, flags, format));
				objectPtr = sharedObject;
			} else {
				staticObject.eat(new StaticObject(objectKey, flags, format));
				objectPtr =staticObject;
			}
			objectPtr->setResolved(filePath + objName->asIdentifier());
		}

		if (line.find("***Object Module Header***") != line.npos) {
			section = ObjectModuleHeader;
			continue;
		}

		if (line.find("***Import File Strings***") != line.npos) {
			section = ImportFileStrings;
			parseRunpathDeplibs(&dumper, sharedObject);
			section = NoSection;
			continue;
		}

		if (section == ObjectModuleHeader && line.find("Flags=(") == 0) {
			line.erase(0, line.find("("));
			line.erase(0, 1);
			line.erase(line.find(")"));
			flags = line;
			continue;
		}

		if (section == ObjectModuleHeader && line.find("Magic = ") == 0) {
			line.erase(0, line.find("("));
			line.erase(0, 1);
			line.erase(line.find(")"));
			format = line;
			continue;
		}

	}
	if (objectPtr) {
		objectCache_[idPath + base + currMember] = objectPtr;
		if (currMember.length() > 0) {
			libraryMemberCache_[idPath+base].push_back(currMember);
		}
		if (member.length() == 0 || member == currMember) {
			out.push_back(objectPtr);
			++ rv;
		}
	}

	if (dumper.close() != 0 && rv == 0) {
		brokenFilesCache_[filePath + base] = true;
	}

	return rv;
}

}
