/* Copyright 2008 Gentoo Foundation
 * Distributed under the terms of the GNU General Public License v2
 * Author: Michael Haubenwallner <haubi@gentoo.org>
 */

#include <string>
#include <vector>
#include <map>
#include <iostream>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "ObjectReader.h"

using namespace aixdll;

typedef ObjectReader::ObjectContainer ObjectContainer;
typedef std::map<std::string, ObjectPtr> SeenContainer;
typedef std::vector<std::string> MissingContainer;

static bool collectSharedObjects(ObjectReader& reader, bool shallow
		, std::string const& path, std::string const& identifier
		, SeenContainer& read, ObjectContainer& resolved, MissingContainer& unread, SearchPath& runpath
	)
{
	ObjectContainer objects;

	if (reader.readObjects(path + identifier, identifier, objects) == 0) {
		return false;
	}

	ObjectContainer::iterator obj;
	for(obj = objects.begin(); obj != objects.end(); obj++) {
		SharedObject* sharedObject = dynamic_cast<SharedObject*>(obj->ptr());
		if (sharedObject == NULL) {
			continue;
		}

		runpath.append(*sharedObject->runpath().ptr());
		read[sharedObject->key()->asIdentifier()] = *obj;
		resolved.push_back(*obj);

		if (shallow) continue;

		SharedObject::DependencyContainer const& dependencies =
			sharedObject->getDependencies();
		SharedObject::DependencyContainer::const_iterator dep;
		for(dep = dependencies.begin(); dep != dependencies.end(); dep++) {
			if ((*dep)->key()) {
				if (read.find((*dep)->key()->asIdentifier()) != read.end()) {
					continue;
				}
				unread.push_back((*dep)->key()->asIdentifier());
			} else
			if ((*dep)->name()) {
				if (read.find((*dep)->name()->asIdentifier()) != read.end()) {
					continue;
				}
				unread.push_back((*dep)->name()->asIdentifier());
			}
		}
	}

	return true;
}

void ldd_r(ObjectReader& reader, bool shallow, std::string const& filename)
{
	SeenContainer read;
	MissingContainer unread;
	ObjectContainer resolved;
	SearchPath runpath;
	char const* envLibpath = getenv("LIBPATH");
	if (envLibpath) {
		runpath.appendSeparated(envLibpath);
	}

	if (!collectSharedObjects(reader, false, "", filename, read, resolved, unread, runpath)) {
		return;
	}

	resolved.clear();

	while(unread.size() > 0) {
		MissingContainer::iterator unr = unread.begin();
		ObjectPathPtr path;
		ObjectNamePtr name;
		std::string unrName = *unr;
		unread.erase(unr);
		if (read.find(unrName) != read.end()) {
			continue;
		}
		splitIdentifier(unrName, path, name);
		if (path) {
			if (!collectSharedObjects(reader, shallow
					, ""
					, unrName
					, read
					, resolved
					, unread
					, runpath
			)) {
				resolved.push_back(
					ObjectPtr(new Object(
						ObjectKeyPtr(new ObjectKey(path, name))
						, "", ""
					))
				);
			}
		} else {
			SearchPath::PathContainer::const_iterator p;
			for(p = runpath.components().begin()
			  ; p != runpath.components().end()
			  ; p++
			) {
				if (collectSharedObjects(reader, shallow
						, (*p)
						, name->asIdentifier()
						, read
						, resolved
						, unread
						, runpath
				)) {
					break;
				}
			}
			if (p == runpath.components().end()) {
				resolved.push_back(
					ObjectPtr(new Object(
						ObjectKeyPtr(new ObjectKey(path, name))
						, "", ""
					))
				);
			}
		}
	}

	for(ObjectContainer::iterator r = resolved.begin()
	  ; r != resolved.end()
	  ; r++
	) {
		std::cout << "	"; // tab
		if ((*r)->getResolved().length() == 0) {
			std::cout
				<< (*r)->key()->asIdentifier()
				<< " => not found"
			;
		} else
		if ((*r)->key()->path()) {
			std::cout << (*r)->getResolved();
		} else {
			std::cout
				<< (*r)->key()->asIdentifier()
				<< " => " << (*r)->getResolved();
		}
		std::cout << std::endl;
	}
}

bool ldd(ObjectReader& reader, bool shallow, MissingContainer const& toread)
{
	ObjectContainer objects;
	bool err = false;

	for(MissingContainer::const_iterator filename = toread.begin()
	  ; filename != toread.end()
	  ; filename++
	) {
		if (reader.readObjects(*filename, *filename, objects) == 0) {
			std::cerr << "ldd: " << *filename << ": " << strerror(ENOENT) << std::endl;
			err = true;
		}
	}

	for(ObjectContainer::iterator o = objects.begin()
	  ; o != objects.end()
	  ; o++
	) {
		SharedObject *sharedObject = dynamic_cast<SharedObject*>(o->ptr());
		if (!sharedObject) {
			continue;
		}
		if (objects.size() > 1) {
			std::cout << (*o)->key()->asIdentifier() << ":" << std::endl;
		}
		ldd_r(reader, shallow, (*o)->key()->asIdentifier());
	}

	return err;
}

int main(int argc, char **argv)
{
	ObjectReader reader;
	int i;
	std::string arg;
	bool shallow = false;
	MissingContainer toread;

	for(i = 1; i < argc; i++) {
		arg = argv[i];
		if (arg == "--shallow") {
			shallow = true;
			continue;
		}
		toread.push_back(arg);
	}

	return ldd(reader, shallow, toread);
}
