/* Copyright 2008 Gentoo Foundation
 * Distributed under the terms of the GNU General Public License v2
 * Author: Michael Haubenwallner <haubi@gentoo.org>
 */

#ifndef _aixdll_smartptr_h
#define _aixdll_smartptr_h

namespace aixdll {

class RefCounted {
public:
   RefCounted() : refcount_(0) {}
   virtual ~RefCounted() {}

   RefCounted(const RefCounted&) : refcount_(0) {}
   RefCounted& operator=(const RefCounted&) { return *this; }

   virtual void ref() { refcount_++; }
   virtual void unref() { if (!--refcount_) delete this; }

   int refcount() const { return refcount_; }

   static void ref(RefCounted* r) { if (r) r->ref(); }
   static void unref(RefCounted* r) { if (r) r->unref(); }

private:
   int refcount_;
};


template <class T> class SmartPtr {
public:

   SmartPtr(): ptr_(0) {}

   explicit SmartPtr(T* p): ptr_(p) { RefCounted::ref(ptr_); }

   template<class RHS>
   SmartPtr(const SmartPtr<RHS>& p): ptr_(0) { operator=(p); }

   SmartPtr(const SmartPtr& p): ptr_(0) { operator=(p); }

   ~SmartPtr() { RefCounted::unref(ptr_); }

   template<class RHS>
   SmartPtr&
   operator=(
      const SmartPtr<RHS>& p)
   {
      if (this->ptr() != p.ptr()) {
         RefCounted::ref(p.ptr());
         RefCounted::unref(this->ptr());
         ptr_ = p.ptr();
      }
      return *this;
   }

   SmartPtr&
   operator=(
      const SmartPtr& p)
   {
      if (this->ptr() != p.ptr()) {
         RefCounted::ref(p.ptr());
         RefCounted::unref(this->ptr());
         ptr_ = p.ptr();
      }
      return *this;
   }

   template<class RHS>
   void eat(RHS* p)
   {
      if (p == ptr_) return;
      RefCounted::unref(ptr_);
      ptr_ = p;
      RefCounted::ref(p);
   }

   void forget()
   {
      if (ptr_) {
         RefCounted::unref(ptr_);
         ptr_ = NULL;
      }
   }

   operator bool() const {
      return static_cast<bool>(ptr_);
   }
   bool operator !() const { return !ptr_; }
   bool operator< (const SmartPtr& that) const { return ptr_< that.ptr_; }
   bool operator==(const SmartPtr& that) const { return ptr_==that.ptr_; }
   bool operator> (const SmartPtr& that) const { return ptr_> that.ptr_; }
   bool operator<=(const SmartPtr& that) const { return ptr_<=that.ptr_; }
   bool operator>=(const SmartPtr& that) const { return ptr_>=that.ptr_; }

   T* ptr() const { return ptr_; }
   T* operator->() { return ptr_; }
   const T* cptr() const { return ptr_; }
   const T* operator->() const { return ptr_; }


private:
   T* ptr_;
};

}

#endif
